﻿module Windows.CVideoRecorder;
import core.thread, std.format, std.path, std.range, std.stream, std.zlib;
import win32.windows;
import DOpenGL.gl;
import Shared.CObeliskException;

struct BITMAPINFOFix
{
	BITMAPINFOHEADER bmiHeader;
	RGBQUAD bmiColors[256];
}

class CVideoRecorder
{
	private string mOutputDirectory;
	private int mXSize;
	private int mYSize;
	private HANDLE mStopEvent;
	private HANDLE mWorkEvent;
	private CRITICAL_SECTION mProcessCS;
	private HANDLE mProcessingEvent;
	private Thread mThread = null;
	private HDC mMemoryDC = null;
	private HBITMAP mBitmap = null;
	private BITMAPINFOFix mInfo;
	private ubyte[] mImageData;
	private int mFrame = 0;
	private int mCount = 0;

	private File mOutStream;

	public this(string aOutputDirectory)
	{
		mOutputDirectory = aOutputDirectory;
		mStopEvent = CreateEvent(null, true, false, null);
		mWorkEvent = CreateEvent(null, false, false, null);
		InitializeCriticalSection(&mProcessCS);
		mProcessingEvent = CreateEvent(null, false, false, null);
	}



	public ~this()
	{
		bool CHResult = 0 != CloseHandle(mStopEvent);
		assert(CHResult);
		CHResult = 0 != CloseHandle(mWorkEvent);
		assert(CHResult);
		DeleteCriticalSection(&mProcessCS);
		CHResult = 0 != CloseHandle(mProcessingEvent);
		assert(CHResult);
		if (mMemoryDC !is null)
		{
			bool DDResult = 0 != DeleteDC(mMemoryDC);
			assert(DDResult);
			bool DOResult = 0 != DeleteObject(mBitmap);
			assert(DOResult);
		}
	}



	public void Start(int aXSize, int aYSize)
	{
		mXSize = aXSize;
		mYSize = aYSize;
		mThread = new Thread(&Run);
		mThread.start();
	}



	public void Stop()
	{
		bool SEResult = 0 != SetEvent(mStopEvent);
		assert(SEResult);
		mThread.join();
	}



	public void GetFrame(HDC aDC)
	{
		mCount = 1 - mCount;
		if (mCount == 0)
			return;
		if (mMemoryDC is null)
		{
			mMemoryDC = CreateCompatibleDC(aDC);
			assert(mMemoryDC !is null);
			mBitmap = CreateCompatibleBitmap(aDC, mXSize, mYSize);
			assert(mBitmap !is null);
		}
		EnterCriticalSection(&mProcessCS);
		HGDIOBJ PreviousBitmap = SelectObject(mMemoryDC, mBitmap);
		assert(PreviousBitmap !is null);
		bool BBResult = 0 != BitBlt(mMemoryDC, 0, 0, mXSize, mYSize, aDC, 0, 0, SRCCOPY);
		assert(BBResult);
		PreviousBitmap = SelectObject(mMemoryDC, PreviousBitmap);
		assert(PreviousBitmap !is null);
		if (mInfo.bmiHeader.biSize == 0)
		{
			mInfo.bmiHeader.biSize = BITMAPINFOHEADER.sizeof;
			bool GDBResult = 0 != GetDIBits(mMemoryDC, mBitmap, 0, mYSize, null,
					cast(BITMAPINFO*)&mInfo, DIB_RGB_COLORS);
			assert(GDBResult);
			mImageData = new ubyte [mInfo.bmiHeader.biSizeImage];
		}
		bool GDBResult = 0 != GetDIBits(mMemoryDC, mBitmap, 0, mYSize, mImageData.ptr,
				cast(BITMAPINFO*)&mInfo, DIB_RGB_COLORS);
		assert(GDBResult);
		LeaveCriticalSection(&mProcessCS);
		bool SEResult = 0 != SetEvent(mWorkEvent);
		assert(SEResult);
		DWORD WFSOResult = WaitForSingleObject(mProcessingEvent, INFINITE);
		assert(WFSOResult == WAIT_OBJECT_0);
	}



	private void Run()
	{
		HANDLE[2] WaitObjects;
		WaitObjects[0] = mStopEvent;
		WaitObjects[1] = mWorkEvent;
		while (true)
		{
			DWORD WFMOResult = WaitForMultipleObjects(2, WaitObjects.ptr, false, INFINITE);
			assert(WFMOResult >= WAIT_OBJECT_0 && WFMOResult <= WAIT_OBJECT_0 + 1);
			if (WFMOResult == WAIT_OBJECT_0)
				return;
			EnterCriticalSection(&mProcessCS);
			bool SEResult = 0 != SetEvent(mProcessingEvent);
			assert(SEResult);

			SaveFrame();

			LeaveCriticalSection(&mProcessCS);
		}
	}



	private void SaveFrame()
	{
		Compress c = new Compress();
		ubyte[] Data;
		int LastOffset = 0;
		for (int y = 0; y < mYSize; y++)
		{
			Data ~= cast(ubyte[])c.compress(cast(void[])[cast(ubyte)0]);
			int Offset = (y + 1) * mXSize * 4;
			// set alpha values to 255
			for (int o = LastOffset + 3; o < Offset; o += 4)
				mImageData[o] = 255;
			Data ~= cast(ubyte[])c.compress(cast(void[])mImageData[LastOffset .. Offset]);
			LastOffset = Offset;
		}
		Data ~= cast(ubyte[])c.flush();

		auto Writer = appender!string();
		formattedWrite(Writer, "Frame_%08d.png", mFrame++);
		string FilePath = join(mOutputDirectory, Writer.data);
		try
		{
			mOutStream = new File(FilePath, FileMode.OutNew);
		}
		catch (OpenException)
		{
			// mOutStream: handle this somehow
			mOutStream = null;
		}
		WriteBytes([137, 80, 78, 71, 13, 10, 26, 10]);
		WriteChunkStart(13, "IHDR");
		WritePNGUInt(mXSize);
		WritePNGUInt(mYSize);
		WriteBytes([8, 6, 0, 0, 0]);
		WriteChunkEnd();
		WriteChunkStart(Data.length, "IDAT");
		WriteBytes(Data);
		WriteChunkEnd();
		WriteChunkStart(0, "IEND");
		WriteChunkEnd();
	}



	private void WriteBytes(ubyte[] aData)
	{
		mOutStream.writeExact(cast(void*)aData, aData.length);
	}



	private void WriteChunkStart(int aLength, string aChunkName)
	in
	{
		assert(aChunkName.length == 4);
	}
	body
	{
		WritePNGUInt(aLength);
		foreach (char s; aChunkName)
			mOutStream.write(cast(ubyte)s);
	}



	private void WriteChunkEnd()
	{
		// TODO: implement this
		WriteBytes([0, 0, 0, 0]);
	}



	private void WritePNGUInt(int aValue)
	in
	{
		assert(aValue >= 0 && aValue <= 2147483647);
	}
	body
	{
		WriteBytes([
			cast(ubyte)((aValue & 0xff000000) >> 24),
			cast(ubyte)((aValue & 0xff0000) >> 16),
			cast(ubyte)((aValue & 0xff00) >> 8),
			cast(ubyte)(aValue & 0xff),
		]);
	}
}